/*$Id: ap.h,v 13.39 1999/05/14 08:53:21 al Exp $  -*- C++ -*-
 * stuff for the "ap" family of parsing functions
 */
#ifndef AP_H
#define AP_H
#include "md.h"
/*--------------------------------------------------------------------------*/
enum AP_MODE{
  ONEPASS,	/* NO */
  REPEAT	/* YES */
};
enum AP_TYPE{
  aDOUBLE,	/* double */
  aUDOUBLE,	/* double, force positive */
  aINT,		/* int */
  aUINT,	/* int, force positive */
  aENUM,	/* int, set to next param, do not read */
  aODOUBLE,	/* double with offset */
  a2DOUBLE,	/* 2 doubles */
  aORENUM,	/* int, or to next param, do not read */
  aANDENUM,	/* int, and to next param, do not read */
  aFUNCTION,	/* call function with params cmd,cnt */
  a2FUNCTION,	/* 2 functions, in sequence */
  aSDOUBLE,	/* double with scale */
  aFINT,	/* int, but take in double format */
  aIDOUBLE,	/* inverted double */
  aSIDOUBLE,	/* scaled, inverted double (scaled first) */
  aOIDOUBLE,	/* offset, inverted double (offset first) */
  aOCTAL,	/* int in octal */
  aHEX,		/* int in hex */
  aFFUNCTION,	/* call function with CS object */
  a2FFUNCTION,	/* 2 functions, in sequence */
  aTRUE,	/* type bool */
  aFALSE,	/* another type bool */
  aIGNORE	/* ignore this keyword */
};
enum AP_MOD{
  mNONE,	/* nothing special */
  mSCALE,	/* scale it after reading */
  mOFFSET,	/* add an offset */
  mINVERT,	/* save 1 / the number */
  mPOSITIVE,	/* store absolute value */
  mOCTAL,	/* read the number in octal */
  mHEX		/* read the number in hex */
};
  
class CS {
private:
  const char* _cmd;
  int  _cnt;
  int  _last;
  bool _ok;
  int  _length;
public:
  explicit    CS(const char* c,int i=0)
    :_cmd(c), _cnt(i), _last(-1), _ok(true), _length(strlen(c)) {}
  int	      cursor()const	{return _cnt;}
  CS&	      reset(int c=0)	{_cnt=c; _ok=true; return *this;}
  const char* fullstring()const	{return _cmd;}
  const char* tail()const	{return &_cmd[_cnt];}
  char	      ctoc()		{return _cmd[_cnt++];}
  CS&	      skip(int c=1)	{_cnt+=c; _ok=_cnt<=_length; return *this;}
  char	      peek()const	{return _cmd[_cnt];}
  bool	      match(char c)const{return (_cmd[_cnt]==c);}
  bool	      match1(const char *c)const{return _cmd[_cnt]&&strchr(c,_cmd[_cnt]);}
  bool        stuck()		{_ok=_last<_cnt; _last=_cnt; return !_ok;}
  bool        gotit()		{_ok=_last<_cnt; return _ok;}
  bool	      more()		{skipbl(); return _cmd[_cnt]!='\0';}
  bool	      end()		{skipbl(); return _cmd[_cnt]=='\0';}
  bool	      is_xdigit()const
		{untested(); return (match1("0123456789abcdefABCDEF"));}
  bool	      is_digit()const	{return (match1("0123456789"));}
  bool	      is_pfloat()const	{return (match1(".0123456789"));}
  bool	      is_float()const	{return (match1("+-.0123456789"));}
  bool	      is_alpha()const	{return !!isalpha(toascii(_cmd[_cnt]));}
  bool	      is_term(const char* t)
		{char c=peek();return (c=='\0' || isspace(c) || strchr(t,c));}
  CS&	      skipcom()		{return skip1b(",=");}
  CS&	      skiplparen()	{return skip1b("([");}
  CS&	      skiprparen()	{return skip1b(")]");}
  CS&	      skipequal()	{return skip1b("=");}
  double      ctopf()		{return fabs(ctof());}
  CS&         operator>>(int&x) {x=ctoi();return *this;}
  CS&         operator>>(unsigned&x){untested(); x=ctou();return *this;}
  CS&         operator>>(double&x){x=ctof();return *this;}
  CS&         warn(int i)	{return warn(i,cursor());}
	      operator bool()const{return _ok;}
  double      ctof();				// ap_ctof.cc
  int	      ctoi();				// ap_ctoi.cc
  unsigned    ctou();
  int	      ctoo();
  int	      ctox();
  char*       ctostr(char*,int,const char*);	// ap_ctos.cc
  CS&	      check(int);			// ap_error.cc
  CS&	      warn(int,int);
  CS&         set(const char*, bool*, bool);	// ap_get.cc
  CS&         set(const char*, int*, int);
  CS&         get(const char*, int*, AP_MOD=mNONE, int=0);
  CS&         get(const char*, double*, AP_MOD=mNONE, double=0);
  CS&         get(const char*, double*, double*);
  CS&	      pmatch(const char*);		// ap_match.cc
  CS&	      match(const char* s) {return pmatch(s);}
  CS&	      scan(const char*);
  int	      argparse(AP_MODE,...);		// ap_parse.cc
  CS&	      skipbl();				// ap_skip.cc
  CS&	      skip1b(const char*);
  CS&	      skip1(const char*);
  CS&	      skiparg();
  CS&	      skipto(const char*);
  CS&	      skipto(char);
};	
/*--------------------------------------------------------------------------*/
template <class T>
inline CS& set(CS& cmd, const char* key, T* val, T newval)
{
  if (cmd.pmatch(key)){
    *val = newval;
  }
  return cmd;
}    
/*--------------------------------------------------------------------------*/
#if defined(PEDANTIC_TEMPLATES) || defined(PedTemP)
template <class T>
inline CS& set(CS& cmd, char* key, T* val, T newval)
{
  untested();
  if (cmd.pmatch(key)){
    *val = newval;
  }
  return cmd;
}
#endif
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
#endif
