/*$Id: c_getckt.cc,v 13.28 1999/04/11 08:25:27 al Exp $ -*- C++ -*-
 * build, get, merge, "<" commands
 * process circuit files, and keyboard entry
 */
#include "d_admit.h"
#include "d_cap.h"
#include "d_cccs.h"
#include "d_ccvs.h"
#include "d_coil.h"
#include "d_coment.h"
#include "d_cs.h"
#include "d_dot.h"
#include "d_diode.h"
#include "d_logic.h"
#include "d_mos.h"
#include "d_res.h"
#include "d_subckt.h"
#include "d_trln.h"
#include "d_vccs.h"
#include "d_vcvs.h"
#include "d_vs.h"
#include "d_switch.h"
#include "ap.h"
#include "declare.h"	/* getcmd */
#include "c_comand.h"
/*--------------------------------------------------------------------------*/
//	void	CMD::build(CS&);
//	void	CMD::get(CS&);
//	void	CMD::merge(CS&);
//	void	CMD::run(CS&);
static  void    getmerge(CS&);
static	CARD* parsebranch(char*,int);
static	CARD* cparse(CS&);
static	CARD* dodot(CS&);
/*--------------------------------------------------------------------------*/
extern char head[];                 /* place to store title line            */
extern run_mode_t run_mode;
/*--------------------------------------------------------------------------*/
/* cmd_build: build command
 * get circuit description direct from keyboard (or stdin if redirected)
 * Command syntax: build <before>
 * Bare command: add to end of list
 * If there is an arg: add before that element
 * null line exits this mode
 * preset, but do not execute "dot cards"
 */
void CMD::build(CS& cmd)
{
  STATUS::get.reset().start();
  //STATUS::iter[iTOTAL] = 1;
  run_mode = rPRESET;
  SIM::uninit();
  CARD_LIST::putbefore = findbranch(cmd, &CARD_LIST::card_list);
  CARD* brh;
  do{
    char buffer[BIGBUFLEN];
    getcmd(">",buffer,BIGBUFLEN);
    brh = parsebranch(buffer,true);
  }while (exists(brh));
  STATUS::get.stop();
}
/*--------------------------------------------------------------------------*/
/* cmd_get: get command
 * get circuit from a file, after clearing the old one
 * preset, but do not execute "dot cards"
 */
void CMD::get(CS& cmd)
{
  untested();
  STATUS::get.reset().start();
  run_mode = rPRESET;
  CS nil("");
  clear(nil);
  getmerge(cmd);
  STATUS::get.stop();
}
/*--------------------------------------------------------------------------*/
/* cmd_merge: merge command
 * as get, but do not clear first
 */
void CMD::merge(CS& cmd)
{
  untested();
  STATUS::get.reset().start();
  run_mode = rPRESET;
  getmerge(cmd);
  STATUS::get.stop();
}
/*--------------------------------------------------------------------------*/
/* cmd_run: "<" and "<<" commands
 * run in batch mode
 * "<<" clears old circuit first, "<" does not
 * get circuit from file, execute dot cards in sequence
 */
void CMD::run(CS& cmd)
{
  STATUS::get.reset().start();
  while (cmd.match('<')){
    CS nil("");
    clear(nil);
    cmd.skip();
    cmd.skipbl();
  }
  run_mode = rEXECUTE;
  getmerge(cmd);
  STATUS::get.stop();
}
/*--------------------------------------------------------------------------*/
/* getmerge: actually do the work for "get", "merge", etc.
 */
static void getmerge(CS& cmd)
{
  static FILE *filen;	/* file number (static to avoid leaks on exceptions)*/

  //STATUS::iter[iTOTAL] = 1;
  SIM::uninit();
  xclose(&filen);
  filen = xopen(cmd,"","r");
  if (!filen)
    error(bERROR, "");
  
  bool  echoon = false;		/* echo on/off flag (echo as read from file) */
  bool  liston = false;		/* list on/off flag (list actual values) */
  bool  quiet = false;		/* don't echo title */
  cmd.stuck();
  do{
    cmd.set("Echo",  &echoon, true);
    cmd.set("List",  &liston, true);
    cmd.set("Quiet", &quiet,  true);
  }while (cmd.more() && !cmd.stuck());
  cmd.check(bWARNING);
  
  char buffer[BIGBUFLEN];
  if (!getlines(buffer, BIGBUFLEN, filen))	/* title */
    error(bWARNING, "empty circuit file\n");
  trim(buffer);
  if (!quiet)
    IO::mstdout << buffer << '\n';
  if (*buffer)
    strcpy(head, buffer);
  
  CARD_LIST::putbefore = CARD_LIST::fat_iterator(&(CARD_LIST::card_list),
					    CARD_LIST::card_list.end());
  while (getlines(buffer, BIGBUFLEN, filen)){
    if (echoon)
      IO::mstdout << buffer << '\n';
    CARD* brh = parsebranch(buffer,false);
    if (liston  &&  exists(brh)){
      brh->print(IO::mstdout, false);
    }
  }
  xclose(&filen);
}
/*--------------------------------------------------------------------------*/
/* parsebranch: parse an input line, and process it
 */
static CARD *parsebranch(char *buffer, int alwaysdupcheck)
{
  CARD_LIST::fat_iterator ci(CARD_LIST::putbefore);
  if (OPT::dupcheck ||  alwaysdupcheck){
    CS buf(buffer);
    ci = findbranch(buf, &CARD_LIST::card_list);
  }
 
  CARD* brh = NULL;
  {if (!ci.isend()){ // found one like it, reparse same one
    brh = *ci;
    error(bWARNING, "replacing [%s]\n", brh->printlabel());
    CS buf(buffer);
    brh->parse(buf);
  }else{
    CS buf(buffer);
    CARD_LIST::fat_iterator before = CARD_LIST::putbefore;
    /* Need to stash where to put it, because a subckt will change it.
     * The subckt card itself needs to be placed at the old location,
     * even though parsing it (cparse) changed it.
     */
    brh = cparse(buf);
    if (exists(brh)){
      before.list()->insert(before, brh);
    }
  }}
  if (exists(brh)){
    if (brh->isdevice()){
      SIM::uninit();
    }
  }
  return brh;
}
/*--------------------------------------------------------------------------*/
/* cparse: circuit parse: parse one line of a netlist
 * mostly, dispatches to the proper function.
 */
static CARD *cparse(CS& cmd)
{
  cmd.skipbl();
  if (cmd.is_digit())
    cmd.ctoi();	/* ignore line numbers */
  
  CARD* brh = NULL;
  int here = cmd.cursor();
  switch (toupper(cmd.peek())){
    case '\0':	/* nothing */			break;
    case '.':	brh = dodot(cmd);		break;
    case '\'':
    case '"':
    case ';':
    case '#':
    case '*':	brh = new DEV_COMMENT;		break;
    case 'A':	cmd.warn(bWARNING);		break;
    case 'B':	cmd.warn(bWARNING);		break;
    case 'C':	brh = new DEV_CAPACITANCE;	break;
    case 'D':	brh = new DEV_DIODE;		break;
    case 'E':	brh = new DEV_VCVS;		break;
    case 'F':	brh = new DEV_CCCS;		break;
    case 'G':	brh = new DEV_VCCS;		break;
    case 'H':	brh = new DEV_CCVS;		break;
    case 'I':	brh = new DEV_CS;		break;
    case 'J':	cmd.warn(bWARNING);		break;
    case 'K':	brh = new DEV_MUTUAL_L;		break;
    case 'L':	brh = new DEV_INDUCTANCE;	break;
    case 'M':	brh = new DEV_MOS;		break;
    case 'N':	cmd.warn(bWARNING);		break;
    case 'O':	cmd.warn(bWARNING);		break;
    case 'P':	cmd.warn(bWARNING);		break;
    case 'Q':	cmd.warn(bWARNING);		break;
    case 'R':	brh = new DEV_RESISTANCE;	break;
    case 'S':	brh = new DEV_VSWITCH;		break;
    case 'T':	brh = new DEV_TRANSLINE;	break;
    case 'U':	brh = new DEV_LOGIC;		break;
    case 'V':	brh = new DEV_VS;		break;
    case 'W':	brh = new DEV_CSWITCH;		break;
    case 'X':	brh = new DEV_SUBCKT;		break;
    case 'Y':	brh = new DEV_ADMITTANCE;	break;
    case 'Z':	cmd.warn(bWARNING);		break;
    default:	cmd.warn(bWARNING);		break;
  }
  cmd.reset(here);
  if (exists(brh)){
    brh->parse(cmd);
  }else if (brh){
    error(bWARNING, "internal error: branch has no type <%s>\n",
    	  cmd.fullstring());
    delete brh;
    brh = NULL;
  }
  return brh;
}
/*--------------------------------------------------------------------------*/
static CARD* dodot(CS& cmd)
{
  cmd.skip();
  if (cmd.pmatch("MODEL")){
    cmd.skiparg(); // skip name
    if      (cmd.pmatch("D"    ))  return new MODEL_DIODE;
    else if (cmd.pmatch("NPn"  ))  cmd.warn(bWARNING);
    else if (cmd.pmatch("PNp"  ))  cmd.warn(bWARNING);
    else if (cmd.pmatch("NJf"  ))  cmd.warn(bWARNING);
    else if (cmd.pmatch("PJf"  ))  cmd.warn(bWARNING);
    else if (cmd.pmatch("NMos" ))  return MODEL_MOS_BASE::new_model(cmd);
    else if (cmd.pmatch("PMos" ))  return MODEL_MOS_BASE::new_model(cmd);
    else if (cmd.pmatch("Logic"))  return new MODEL_LOGIC;
    else if (cmd.pmatch("SW"   ))  return new MODEL_SWITCH;
    else if (cmd.pmatch("CSW"  ))  return new MODEL_SWITCH;
    else			   cmd.warn(bWARNING);
  }else if (cmd.pmatch("SUBCKT")){
    return new MODEL_SUBCKT;
  }else{
    return new DEV_DOT;	
  }
  return NULL;
}
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
