/*$Id: d_switch.cc,v 13.32 1999/04/25 07:12:45 al Exp $ -*- C++ -*-
 * voltage (and current) controlled switch.
 * netlist syntax:
 * device:  Sxxxx n+ n- vc+ vc- mname <on>|<off> <model-card-args>
 * model:   .model mname SW <args>
 * current controlled switch
 * device:  Wxxxx n+ n- controlelement mname <on>|<off> <model-card-args>
 * model:   .model mname CSW <args>
 */
#include "ap.h"
#include "d_switch.h"
/*--------------------------------------------------------------------------*/
//		MODEL_SWITCH::MODEL_SWITCH(const char*);
//	void	MODEL_SWITCH::parse(CS&);
//	void	MODEL_SWITCH::print(int,int)const;
//		SWITCH_BASE::SWITCH_BASE();
//		SWITCH_BASE::SWITCH_BASE(const SWITCH_BASE& p);
//	void	SWITCH_BASE::parse_sb(CS&,int);
//	void	SWITCH_BASE::print(int,int)const;
//	void	SWITCH_BASE::expand_sb();
//	bool	SWITCH_BASE::dotr();
//	void	SWITCH_BASE::doac();
//	void	DEV_CSWITCH::expand();
/*--------------------------------------------------------------------------*/
static SWITCH_COMMON Default_SWITCH(CC_STATIC);
/*--------------------------------------------------------------------------*/
MODEL_SWITCH::MODEL_SWITCH()
{
  vt = 0.;
  vh = 0.;
  ron = 1.;
  roff = 1e12;
  type = VOLTAGE;
}
/*--------------------------------------------------------------------------*/
CARD& MODEL_SWITCH::parse(CS& cmd)
{
  cmd.skiparg();		// skip known ".model"
  cmd.ctostr(label, LABELEN, TOKENTERM);
  cmd.stuck();
  /**/ set(cmd, "SW",  &type, VOLTAGE)
    || set(cmd, "CSW", &type, CURRENT);
  if (cmd.stuck()){
    untested();
    cmd.warn(bWARNING);
  }
  cmd.skiplparen();
  cmd.stuck();
  do{
    cmd.get("VT",   &vt);
    cmd.get("VH",   &vh,  mPOSITIVE);
    cmd.get("IT",   &vt);
    cmd.get("IH",   &vh,  mPOSITIVE);
    cmd.get("RON",  &ron);
    cmd.get("ROFF", &roff);
  }while (cmd.more() && !cmd.stuck());
  cmd.skiprparen();
  cmd.check(bWARNING);
  return *this;
}
/*--------------------------------------------------------------------------*/
const CARD& MODEL_SWITCH::print(OMSTREAM where, int)const
{
  where.setfloatwidth(7);
  switch (type){
  case VOLTAGE:
    where << ".model " << label << " sw ("
	  <<   "vt=" << vt
	  << "  vh=" << vh;
    break;
  case CURRENT:
    where << ".model " << label << " csw ("
	  <<   "it=" << vt
	  << "  ih=" << vh;
    break;
  }
  where << "  ron="  << ron
	<< "  roff=" << roff
	<< ")\n";
  return *this;
}
/*--------------------------------------------------------------------------*/
SWITCH_BASE::SWITCH_BASE()
{
  attach_common(&Default_SWITCH);
  inputlabel[0] = '\0';
  input = NULL;
  ic = current_state = previous_state = UNKNOWN;
}
/*--------------------------------------------------------------------------*/
SWITCH_BASE::SWITCH_BASE(const SWITCH_BASE& p):ELEMENT(p)
{
  untested();
  strcpy(inputlabel,p.inputlabel);
  input = NULL;
  ic = p.ic;
  current_state = p.current_state;
  previous_state = p.previous_state;
}
/*--------------------------------------------------------------------------*/
SWITCH_BASE& SWITCH_BASE::parse_sb(CS& cmd, int numnodes)
{
  const SWITCH_COMMON* cc = dynamic_cast<const SWITCH_COMMON*>(common);
  assert(cc);
  SWITCH_COMMON* c = new SWITCH_COMMON(*cc);
  assert(c);

  parselabel(cmd);
  parsenodes(cmd, numnodes);
  if (numnodes == IPRINTNODES){		// if current controlled
    cmd.ctostr(inputlabel, LABELEN, TOKENTERM);
    inputlabel[0] = toupper(inputlabel[0]);
  }
  char modelname[LABELEN+1];
  cmd.ctostr(modelname, LABELEN, TOKENTERM);
  c->set_modelname(modelname);
  assert(!c->has_model());
  cmd.stuck();
  /**/ ::set(cmd, "OFF",    &ic, OFF)
    || ::set(cmd, "ON",	    &ic, ON)
    || ::set(cmd, "UNKNOWN",&ic, UNKNOWN);
  if (cmd.stuck()){
    cmd.check(bWARNING);
  }else{
    untested();
  }
  attach_common(c);
  return *this;
}
/*--------------------------------------------------------------------------*/
const CARD& SWITCH_BASE::print(OMSTREAM where, int)const
{
  const SWITCH_COMMON* c = dynamic_cast<const SWITCH_COMMON*>(common);
  assert(c);
  printlabel(where);
  printnodes(where, numnodes());
  {if (input){
    where << "  ";
    input->printlabel(where);
  }else{
    where << "  " << inputlabel;
  }}
  where << ' ' << c->modelname();

  switch (ic){
    case OFF:	   where << " off";	break;
    case ON:	   where << " on";	break;
    case UNKNOWN: 			break;
  }
  where << '\n';
  return *this;
}
/*--------------------------------------------------------------------------*/
SWITCH_BASE& SWITCH_BASE::expand_sb()
{
  const MODEL_SWITCH* m = dynamic_cast<const MODEL_SWITCH*>(attach_model());
  if (!m){
    untested();
    assert(common);
    error(bERROR, "%s: model %s is not a switch (SW or CSW)\n", 
	  printlabel(), common->modelname());
  }

  setvalue(m->ron);
  y0.f0 = LINEAR;
  y0.f1 = (ic == ON) ? m->ron : m->roff;	/* unknown is off */
  previous_state = current_state = ic;
  m0.f1 = 1./y0.f1;
  m0.c0 = 0.;
  ev    = y0.f1;
  acg   = m0.f1;
  assert(loss == 0.);
  assert(!constant()); /* depends on input */
  return *this;
}
/*--------------------------------------------------------------------------*/
bool SWITCH_BASE::dotr()
{
  assert(!isbypassed());

  const SWITCH_COMMON* c = dynamic_cast<const SWITCH_COMMON*>(common);
  assert(c);
  const MODEL_SWITCH* m = dynamic_cast<const MODEL_SWITCH*>(c->model());
  assert(m);

  assert(STATUS::iter[iSTEP] != 0);
  if (STATUS::iter[iSTEP] <= 1){
    if (SIM::phase == pINIT_DC){   
      y0.f1 = ((ic == ON) ? m->ron : m->roff);	/* unknown is off */
      current_state = ic;
      m0.f1 = 1./y0.f1;
    }
    previous_state = current_state;
  }
  
  y0.x = (input)			/* y0.x is controlling value */
    ? CARD::probe(input,"I")		/* current controlled */
    : n[IN1].v0() - n[IN2].v0();	/* voltage controlled */

  state_t new_state;
  {if (y0.x > m->vt + m->vh){
    new_state = ON;
  }else if (y0.x < m->vt - m->vh){
    new_state = OFF;
  }else{
    new_state = previous_state;
  }}

  {if (new_state != current_state){
    y0.f1 = (new_state == ON) ? m->ron : m->roff;	/* unknown is off */
    current_state = new_state;
    m0.f1 = 1./y0.f1;
    SIM::loadq.push(this);
    store_values();
    set_not_converged();
  }else{
    //assert(y1 == y0);
    set_converged();
  }}
  return converged();
}
/*--------------------------------------------------------------------------*/
CARD& SWITCH_BASE::doac()
{
  ev  = y0.f1;
  acg = m0.f1;
  return acload_passive();
}
/*--------------------------------------------------------------------------*/
CARD& DEV_CSWITCH::expand()
{
  input = dynamic_cast<ELEMENT*>(find_in_scope(inputlabel));
  if (!input){
    untested();
    error(bERROR,"%s: %s cannot be used as input\n",printlabel(),inputlabel);
  }
  return expand_sb();
}
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
