/*$Id: m_cpoly.h,v 13.43 1999/06/01 15:25:51 al Exp $ -*- C++ -*-
 * structs for fixed order poly parameter passing
 * fpoly is by function -- f0 = eval of function
 *			   f1 = 1st derivative
 *			   f2 = 2nd derivative
 *			   etc.
 * cpoly is by series -- c0 = coeff of x^0 term (constant)
 *			 c1 = coeff of x^1 term (1st derivative)
 *			 c2 = coeff of x^2 term
 *			 etc.
 */
#ifndef M_CPOLY_H
#define M_CPOLY_H
#include "io_trace.h"
/*--------------------------------------------------------------------------*/
struct fpoly1_t{	/* first order polynomial	*/
  double  x;		/* the argument			*/
  double  f0;		/* the function (c0 + x*f1)	*/
  double  f1;		/* the first derivative		*/
  explicit fpoly1_t(): x(0), f0(0), f1(0) {}
  //explicit fpoly1_t(const fpoly1_t& p): x(p.x), f0(p.f0), f1(p.f1) {}
  explicit fpoly1_t(double X,double F0, double F1): x(X), f0(F0), f1(F1) {}
  bool operator==(const fpoly1_t& p)const
				{return (f1==p.f1 && f0==p.f0 && x==p.x);}
  fpoly1_t& operator*=(double s){f0*=s; f1*=s; return *this;}
  fpoly1_t& operator+=(const fpoly1_t& s)
				{untested(); f0+=s.f0; f1+=s.f1; return *this;}
};
/*--------------------------------------------------------------------------*/
struct cpoly1_t{	/* first order polynomial	*/
  double  x;		/* the argument			*/
  double  c0;		/* f(x) - x*f'(x), or f0 - x*f1 */
  double  f1;		/* the first derivative		*/
  explicit cpoly1_t(): x(0), c0(0), f1(0) {}
  //explicit cpoly1_t(const cpoly1_t& p): x(p.x), c0(p.c0), f1(p.f1) {}
  bool operator==(const cpoly1_t& p)const
			{untested(); return (f1==p.f1 && c0==p.c0 && x==p.x);}
};
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
#endif
