/*$Id: md.h,v 13.36 1999/05/07 08:37:41 al Exp $ -*- C++ -*-
 * Machine dependent, configuration, and standard includes
 */
#ifndef MD_H_INCLUDED
#define MD_H_INCLUDED
/*--------------------------------------------------------------------------*/
/* system dependent patches needed before the standard includes */
#if defined (WIN32)
#pragma warning (disable: 4786) // id truncated to 255 characters
#pragma warning (disable: 4800) // forcing ___ to bool (performance warning)
#endif
/*--------------------------------------------------------------------------*/
/* c standard collection of includes */
#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <limits.h>		/* limits must precede float (SGI) */
#include <float.h>
#include <math.h>
#include <memory.h>
#include <setjmp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

/* std c++ collection of includes */
#include <string>
#include <complex>
#include <new.h>
#include <vector>
#include <list>
#include <stack>
#include <queue>
#include <algorithm>
#include <map>
/*--------------------------------------------------------------------------*/
#if defined(UNIX)
/* usual but non-standard collection of includes */
#include <unistd.h>		/* some file functions */
#include <sys/param.h>		/* MAXPATHLEN */
#include <sys/resource.h>	/* getrusage, used for timer  */
/* file names, etc. */
#define	ENDDIR		"/"
#define PATHSEP		':'
#define SYSTEMSTARTFILE	"acs.rc"
#define SYSTEMSTARTPATH	getenv("PATH")
#define USERSTARTFILE	".acsrc"
#define	USERSTARTPATH	getenv("HOME")
#define STEPFILE   	"/tmp/SXXXXXX"
#define PLOTFILE    	"acs.plot"
#define HELPFILE    	"acs.hlp"
#define HELPPATH	getenv("PATH")
/* machine and compiler patches */
#define CONST const
#if defined(__GNUG__)
  #define MANUAL_TEMPLATES
  #define std
#else
  #define LINK_TEMPLATES
#endif
#if defined(__hppa__)
  //last test ver = 0.21
  extern "C" int  getrusage(int,struct rusage*);
  //#include <sys/syscall.h>
  //inline int getrusage(int a, struct rusage* b){syscall(SYS_GETRUSAGE,a,b);}
#elif defined(SGI_CC)
  //last test ver = 0.21
  inline double pow(double x, int e){return pow(x,(double)e);}
#elif defined(__sun__)
  //last test ver = 0.21
  extern "C" int  getrusage(int,struct rusage*);
  #if !defined(__GNUG__)
    #define SIGNALARGS int,...
  #endif 
#elif defined(__NeXT__)
//#define _POSIX_SOURCE
  #include <sys/fcntl.h>
  #define INLINE_MATH
#endif
/*--------------------------------------------------------------------------*/
#elif defined(WIN32)
/* usual but non-standard collection of includes */
#include <windows.h>
#include <io.h>
#include <direct.h>
#include <time.h>
/* file names, etc. */
#define	ENDDIR		"/\\"
#define PATHSEP		';'
#define SYSTEMSTARTFILE	"acs.ini"
#define SYSTEMSTARTPATH	getenv("PATH")
#define USERSTARTFILE	"acs.ini"
#define	USERSTARTPATH	getenv("HOME")
#define STEPFILE   	"/tmp/SXXXXXX"
#define PLOTFILE    	"acs.plot"
#define HELPFILE    	"acs.hlp"
#define HELPPATH	getenv("PATH")
/* machine and compiler patches */
#define MANUAL_TEMPLATES
#define NEEDS_GETRUSAGE
#define NEW_IS_BROKEN
#define SIGNALARGS int
#undef  CONST
#define CONST
#define F_OK (00)		/* io.h */
#define X_OK (01)
#define W_OK (02)
#define R_OK (04)
#define MAXPATHLEN (1024)
#define RUSAGE_SELF	0
inline strcasecmp(const char a[], const char b[]){
  return stricmp(a,b);
}
const char __FUNCTION__[] = "";
struct	rusage {
  struct timeval ru_utime;	/* user time used */
  struct timeval ru_stime;	/* system time used */
};
/*--------------------------------------------------------------------------*/
#else
#error what kind of system is this??
#endif
/*--------------------------------------------------------------------------*/
/* my standard collection of includes */
#include "io_error.h"
#include "io_trace.h"
#include "io_.h"

/* constants related to memory size, word size, etc */
enum {
  BUFLEN = 256,
  BIGBUFLEN = 2048,
  MAXWIDTH = 512,
  RECURSE = 8,
  LABELEN = 15
};
/*--------------------------------------------------------------------------*/
typedef std::complex<double> COMPLEX;

#if defined(NDEBUG)
  #define DYNAMIC_CAST static_cast
  // check not needed.  faster.  debug version checks and asserts.
#else
  #define DYNAMIC_CAST dynamic_cast
#endif

#if defined(HAS_EXP_BUG)
  inline double Exp(double x){return (x>-200.) ? exp(x) : 0.;}
#else
  inline double Exp(double x){return exp(x);}
#endif

#if defined(NEEDS_GETRUSAGE)
int getrusage(int,struct rusage*);
#endif

#if !defined(SIGNALARGS)
  #define SIGNALARGS int
#endif
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
#endif
